/*+
    Name:	vgatest2.c
    Date:	26-Aug-1988
    Author:	Kent J. Quirk
    Abstract:	Contains hardware-level tests for VGA Compatibility Suite.
-*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <stdarg.h>
#include <conio.h>
#include "vgabios.h"
#include "vgatest.h"
#include "vgaregs.h"

extern VIDSTAT savemode;

long timer = 0;
/**** s e t t i m e r ****
    Abstract:   This routine starts a timer for a given number of ticks. 
                It waits for the start of a tick, then stores the time of
                the last tick in a register.  The timeron() function
                determines if the timer is still active.
    Parameters: int time -- the number of ticks in the timer
    Returns:    Nothing
****************************/
void settimer(int time)
{
    long now;
    _bios_timeofday(_TIME_GETCLOCK, &now);
    do 
    {				/* wait for start of a tick */
	_bios_timeofday(_TIME_GETCLOCK, &timer);
    } while (timer != now);
    
    timer += time;
}

/**** t i m e r o n ****
    Abstract:	Returns true if timer is on.
    Parameters: Nothing
    Returns:    1 if timer is on, 0 if time is up or timer wasn't started.
                Resets timer if time has expired.
****************************/
int timeron()
{
    long now;
    if (timer != 0)
    {	 
	_bios_timeofday(_TIME_GETCLOCK, &now);
	if (now > timer)
	{
	    timer = 0;
	    return(0);
	}
    }
    return(1);
}

/* 
    The following macro waits for either a condition or a given amount of
    time.  It clarifies the code below.
*/
int waitfor = 0;
#define WAITFOR(cond, time) { waitfor=0; settimer(time); while (timeron()) \
	    if (cond) { waitfor=1; break; } }

/**** t e s t _ 1 _ g e n e r a l _ r e g s ****
    Abstract:   Check the VGA's general register set for one mode by 
                comparing their values to IBM's for the same mode.
    Parameters: int mi -- the mode index of the current mode
    Returns:    Nothing
****************************/
void test_1_general_regs(int mi)
{
    GENERAL_REGS gr;
    int i;
    
    writestring(NORMAL, 0, 0, 0, "General Registers Test - Mode %#x", 
	valid_modes[mi].video_mode);
    read_all_general_regs(&gr);
    compare_data(CMP_GENERAL, mi, &gr);
}

static int int_count = 0, vint_count=0;
void (interrupt far *intsave)();
/**** v e r t i c a l _ i n t _ h a n d l e r ****
    Abstract:   This is an interrupt-level routine designed to handle
                vertical interrupts from those VGA boards that support 
                them.  If it gets control, it checks to see if the VGA
                generated the interrupt.  If it did, it clears the VGA 
                and increments a global counter.  If not, it passes the
                interrupt along to the next device on the interrupt
                chain.  It counts both the total number of interrupts,
                and the number that were from the VGA.
    Parameters: All the chip registers are dummy parameters to this
                routine.  It isn't a good idea to modify any of them 
                in a hardware interrupt handler.
    Returns:    Nothing -- see above.
****************************/
void interrupt far vertical_int_handler(es,ds,di,si,bp,sp,bx,dx,cx,ax,
                                        ip,cs,flags)
unsigned int es,ds,di,si,bp,sp,bx,dx,cx,ax,ip,cs,flags;
{
    BYTE c;
    
    int_count++;
    if (read_general_reg(INPUT_STATUS_0) & 0x80)
    {
	vint_count++;
	c = read_crtc_reg(V_RETRACE_END) & ~VRET_CLR;
	write_crtc_reg(V_RETRACE_END, c);
	write_crtc_reg(V_RETRACE_END, c | VRET_CLR);
	outp(0x20, 0x20);			/* EOI */
    }
    else
        _chain_intr(intsave);               /* not ours; pass it along */
}

/**** t e s t _ g e n e r a l _ r e g s ****
    Abstract:	For each valid mode, this reads all the general registers
		and checks to see that they agree with those defined for
		the IBM Motherboard VGA.  Input Status Register 1 is 
                checked to see that its status bits vary appropriately.
                The vertical retrace interrupt is checked to see if it
                works properly.
    Parameters: None
    Returns:    Nothing
****************************/
void test_general_regs()
{
    BYTE x, lastx=0xff;
    BYTE c;
    int ntoggles = 0;
	
    for_each_mode(test_1_general_regs, QUICK_PAUSE, ALL_MODES);
    
    x = read_general_reg(INPUT_STATUS_1);
    WAITFOR( (((x ^ read_general_reg(INPUT_STATUS_1)) & 0x01) != 0), 1 SEC );
    if (!waitfor)
	error("Display Enable Not bit in ISR1 does not toggle.\n");
    
    WAITFOR( (((x ^ read_general_reg(INPUT_STATUS_1)) & 0x03) != 0), 1 SEC );
    if (!waitfor)
	error("Vertical Retrace bit in ISR1 does not toggle.\n");
	
    settimer(9);	    /* 1/2 second */
    do {
	wait_for_vertical(0);
	ntoggles++;
    } while (timeron());
    if ((ntoggles > 40) || (ntoggles < 35))
	error("Expected vertical retrace 35-40 times in 1/2 sec; got %d.\n",
		ntoggles);
    
    intsave = _dos_getvect(VRET_INT);
    _dos_setvect(VRET_INT, vertical_int_handler);
    c = (read_crtc_reg(V_RETRACE_END) & ~VRET_ENABLE) | VRET_CLR;
    write_crtc_reg(V_RETRACE_END, c);
    settimer(9);	    /* 1/2 second */
    while (timeron())
	;
    write_crtc_reg(V_RETRACE_END, c | VRET_ENABLE);
    _dos_setvect(VRET_INT, intsave);
    if ((vint_count > 40) || (vint_count < 35))
	error("Expected vertical interrupt 35-40 times in 1/2 sec; got %d.\n",
		vint_count);
}

/**** t e s t _ d a c _ r e g s ****
    Abstract:	This reads all the DAC registers, compares their default
		values with those used by IBM, then sets each one, reads 
		them all back, and confirms the accuracy.  The default
		values are reloaded at the end.  Autoincrement and the
		status register are checked.
    Parameters: None   
    Returns:    Nothing
    Comments:
****************************/
void test_dac_regs()
{
    static DAC_REGS dr;
    int i, j;
    void test_1_graphics_io();
    
    GetStatus(&savemode);
    
    SetMode(valid_modes[COLOR256].video_mode);
    test_1_graphics_io(COLOR256);	    /* draw the colors */
    
    write_dac_reg(DAC_PEL_ADDR_R, 0);
    for (i=0; i<256; i++)		/* tests autoincrement and data */
    {
	dr.col[i].red = read_dac_reg(DAC_PEL_DATA);
	dr.col[i].blu = read_dac_reg(DAC_PEL_DATA);
	dr.col[i].grn = read_dac_reg(DAC_PEL_DATA);
    }
    if ((read_dac_reg(DAC_STATE) & 3) != 3)
	error("DAC Status Register reported wrong status after a read.\n");
    
    compare_data(CMP_DAC, 0, &dr);	/* pretend it's mode 0 */
    pausekey(1 SEC);
    
    write_dac_reg(DAC_PEL_ADDR_W, 0);
    for (i=0; i<256; i++)		/* tests autoincrement and data */
    {
	int I = i >> 6;
	write_dac_reg(DAC_PEL_DATA, (I==0 || I==1) ? (BYTE)i & 0x3f : 0);
	write_dac_reg(DAC_PEL_DATA, (I==0 || I==2) ? (BYTE)i & 0x3f : 0);
	write_dac_reg(DAC_PEL_DATA, (I==0 || I==3) ? (BYTE)i & 0x3f : 0);
    }
    if ((read_dac_reg(DAC_STATE) & 3) != 0)
	error("DAC Status Register reported wrong status after a write.\n");
    
    pausekey(2 SEC);
    write_dac_reg(DAC_PEL_ADDR_W, 0);
    for (i=0; i<256; i++)		/* tests autoincrement and data */
    {
	write_dac_reg(DAC_PEL_DATA, dr.col[i].red);
	write_dac_reg(DAC_PEL_DATA, dr.col[i].blu);
	write_dac_reg(DAC_PEL_DATA, dr.col[i].grn);
    }
    
    pausekey(1 SEC);
    SetMode(savemode.video_mode);
}

/**** t e s t _ s e q u e n c e r _ r e g s ****
    Abstract:	This tests sequencer registers:
		Clocking mode:
			Video Disable
                        Map Mask Register is tested with the graphics 
                        controller.
		Char map select
			Repeat BIOS char map select function.
		Memory Mode
			Chain 4, Odd/Even - tested with graphics controller
			Ext. Memory - should be 1
    Parameters: None   
    Returns:    Nothing
    Comments:	The reset register is ignored.
****************************/
void test_sequencer_regs()
{
    BYTE v;
    int x, y;
    
    writestring(NORMAL, 0, 10, 10, "Video will now turn off for 2 seconds.");
    pausekey(2 SEC);
    v = read_sequencer_reg(SQ_CLK_MODE);
    write_sequencer_reg(SQ_CLK_MODE, v | 0x20);
    pausekey(2 SEC);
    write_sequencer_reg(SQ_CLK_MODE, v);

    if ((read_sequencer_reg(SQ_MEM_MODE) & 2) != 2)
	error("Extended memory bit in sequencer is not set.\n");
}

/**** t e s t _ 1 _ c r t c _ r e g s ****
    Abstract:   Checks the CRTC registers for one mode by comparing them
                to the values on the IBM.
    Parameters: The mode index for the current mode.
    Returns:    Nothing
****************************/
void test_1_crtc_regs(int mi)
{
    CRTC_REGS cr;
    int i;
    
    writestring(NORMAL, 0, 0, 0, "CRTC Register Test - Mode %#x", 
	    valid_modes[mi].video_mode);
    read_all_crtc_regs(&cr);
    compare_data(CMP_CRTC, mi, &cr);
}

/**** t e s t _ c r t c _ r e g s ****
    Abstract:	For all modes, set mode, read all regs, compare to IBM.
    Parameters: None   
    Returns:    Nothing
****************************/
void test_crtc_regs()
{
    void test_1_cursor_operation(int m);
    
    for_each_mode(test_1_crtc_regs, QUICK_PAUSE, ALL_MODES);
    GetStatus(&savemode);
    SetMode(savemode.video_mode);
    test_1_cursor_operation(1); 	/* test cursor, check hardware */
    SetMode(savemode.video_mode);
}

/**** w r i t e _ m o d e 0 _ S R ****
    Parameters:     x and y coordinates of point to be drawn
****************************/
void write_mode0_SR (int x, int y)
{
    BYTE    color = (BYTE) ((x + y) % valid_modes[COLOR16].num_colors);
    WORD    offset = (x / 8) + (y * (valid_modes[COLOR16].num_xpixels / 8));
    
    /*
	Write mode 0 with Set / Reset enabled writes the bit contained
	in the Set / Reset register for that plane into each of the eight
	pixels in the byte, IF Set / Reset is enabled for that plane.  We
	will write by setting the Set / Reset Register to the color and
	setting Enable Set / Reset for all planes.
    */
	
    write_graph_reg (MODE, read_graph_reg (MODE) & ~WRITE_MODE_MASK);
    write_graph_reg (SET_RESET, color);
    write_graph_reg (ENABLE_SET_RESET, 0x0F);
    write_graph_reg (BIT_MASK, ((BYTE) 0x80 >> (x % 8)));
    
    /*
	All writing which uses a non-zero value in the Bit Mask register
	must preload the latches by reading the location before writing it.
    */
    
    valid_modes[COLOR16].buf_address[offset] &= 0xFF;
    
    return;
}

/**** w r i t e _ m o d e 0 ****
    Parameters:     x and y coordinates of point to be drawn
****************************/
void write_mode0 (int x, int y)
{
    BYTE    color = (BYTE) ((x + y) % valid_modes[COLOR16].num_colors);
    WORD    offset = (x / 8) + (y * (valid_modes[COLOR16].num_xpixels / 8));
    BYTE    read_value;
    
    /*
	Write mode 0 writes the processor data byte to each of the four
	bitplanes, except those protected by the map mask register.  To
	do what we want, we set the map mask to the color, set the bit
	mask to select only the pixel of interest, and write 0xFF.
    */
    
    write_graph_reg (MODE, read_graph_reg (MODE) & ~WRITE_MODE_MASK);
    write_sequencer_reg (SQ_MAP_MASK, color);
    write_graph_reg (BIT_MASK, ((BYTE) 0x80 >> (x % 8)));
    
    /*
	All writing which uses a non-zero value in the Bit Mask register
	must preload the latches by reading the location before writing it.
    */
    
    read_value = valid_modes[COLOR16].buf_address[offset];
    valid_modes[COLOR16].buf_address[offset] = (BYTE) 0xFF;
    
    return;
}

/**** t e s t _ 1 _ g r a p h i c s _ r e g s ****
    Parameters:     mi is an index into valid_modes[].

****************************/
void test_1_graphics_regs()
{
    int     x, y;
    int     lasty;
    
    /*
	For every pixel on the screen, write that pixel in all write modes
	and check for the writing by reading it back in each of the read
	modes.	The color used at each pixel position on the display is the
	least significant n bits of (x + y), so the read routine can tell
	what the right answer's supposed to be.
    */
    
    SetMode (valid_modes[COLOR16].video_mode);
    lasty = (fast) ? 10 : valid_modes[COLOR16].num_ypixels;
    for (y = 0; y < lasty; y++)
    {	 
	for (x = 0; x < valid_modes[COLOR16].num_xpixels; x++)
	{
	    write_mode0 (x, y);
	    /* read_mode0 (x, y); */
	}
	if (kbhit())
	{
	    getch();
	    break;
	}
    }
    lasty = y;
    
    SetMode (valid_modes[COLOR16].video_mode);
    for (y = 0; y < lasty; y++)
	for (x = 0; x < valid_modes[COLOR16].num_xpixels; x++)
	{
	    write_mode0_SR (x, y);
	    /* read_mode0 (x, y); */
	}			  
/*
    for (y = 0; y < lasty; y++)
	for (x = 0; x < valid_modes[COLOR16].num_xpixels; x++)
	{
	    write_mode0 (x, y);
	    read_mode0 (x, y);
	}
    for (y = 0; y < lasty; y++)
	for (x = 0; x < valid_modes[COLOR16].num_xpixels; x++)
	{
	    write_mode0 (x, y);
	    read_mode0 (x, y);
	}
	    write_mode0_SR (x, y, mi);
	    read_mode0_C4 (x, y, mi);
	    write_mode1 (x, y, mi);
	    read_mode1 (x, y, mi);
	    write_mode2 (x, y, mi);
	    read_mode1 (x, y, mi);
	    write_mode3 (x, y, mi);
	    read_mode0 (x, y, mi);
	}			  */
    
    return;
}

/**** t e s t _ g r a p h i c s _ r e g s ****
    Abstract:	    For 640 x 350 16-color mode (or monochrome, if needed):
    
    Parameters:
    Returns:
    Comments:
****************************/
void test_graphics_regs()
{   
    GetStatus(&savemode);
    test_1_graphics_regs ();
    SetMode(savemode.video_mode);
}

/**** s e t _ l i n e _ c o m p a r e ****
    Abstract:   Used in Pan/Scroll test, this sets the aggregate line 
                compare register.  It is a 10-bit register, with
                the lowest bits in the CRTC Line Compare register, the
                next bit in the Overflow register, and the last bit in the
                Max Scan Line register.  We can thank backward compatibility
                for this.
    Parameters: The new value of the line compare register.
    Returns:    Nothing.
    Comments:
****************************/
void set_line_compare(int v)
{
    BYTE t;
    write_crtc_reg(LINE_COMPARE, (BYTE)(v & 0xFF));
    v >>= 8;
    t = read_crtc_reg(OVERFLOW) & (~0x10);
    if (v & 1)
	t |= 0x10;
    write_crtc_reg(OVERFLOW, t);
    v >>= 1;
    t = read_crtc_reg(MAX_SCAN_LINE) & (~0x40);
    if (v & 1)
	t |= 0x40;
    write_crtc_reg(MAX_SCAN_LINE, t);
}

/**** s e t _ p a n _ c o m p a t ****
    Abstract:   This sets the Pan Compatibility bit in the Attribute
                Mode Control register.  This prevents the bottom half
                of the screen from jumping around while we attempt to
                do horizontal scrolling.
    Parameters: None
    Returns:    Nothing
****************************/
void set_pan_compat()
{
    BYTE t;
    t = read_attrib_reg(ATTR_MODE_CTRL);
    t |= 0x20;		/* PEL Panning Compatibility bit */
    write_attrib_reg(ATTR_MODE_CTRL, t);
}

/**** r e s e t _ p a n _ c o m p a t ****
    Abstract:   Undoes what set_pan_compat() did.
****************************/
void reset_pan_compat()
{
    BYTE t;
    t = read_attrib_reg(ATTR_MODE_CTRL);
    t &= ~0x20; 	/* PEL Panning Compatibility bit */
    write_attrib_reg(ATTR_MODE_CTRL, t);
}

/**** g e t _ s t a r t _ a d d r ****
    Abstract:   Read the CRTC Start Address registers and return
                a 16-bit value.
    Parameters: None
    Returns:    WORD value of the two registers together.
****************************/
WORD get_start_addr()
{
    WORD x;
    
    x = read_crtc_reg(START_ADDR_LO);
    x += read_crtc_reg(START_ADDR_HI) << 8;
    return(x);
}

/**** s e t _ s t a r t _ a d d r ****
    Abstract:   Sets the CRTC Start Address registers to a WORD.
    Parameters: The new WORD value of the register pair.
    Returns:    Nothing
****************************/
void set_start_addr(WORD x)
{
    write_crtc_reg(START_ADDR_LO, (BYTE)(x & 0xff));
    write_crtc_reg(START_ADDR_HI, (BYTE)(x >> 8));
}

/**** s e t _ s c r n _ o r g ****
    Abstract:   Sets the screen origin for the top half of the screen.
    Parameters: It takes the current mode index (so it can understand the
                layout of the memory map), the video page number to use,
                and the x and y position on that page of the upper left
                corner of the screen.
    Returns:    Nothing.
****************************/
void set_scrn_org(int mi, BYTE page, WORD x, WORD y)
{
    WORD startaddr;
    BYTE startline, startcol;
    BYTE bpp;
    int charwidth = 8;
    int rowheight;
    int scrnwidth = valid_modes[mi].num_cols;
    
    bpp = read_sequencer_reg(SQ_CLK_MODE) & 1;      /* bits per pixel */
    if (bpp)
	charwidth = 8;
    else
	charwidth = 9;
	
    if (valid_modes[mi].isgraphics)
	rowheight = valid_modes[mi].num_ypixels/valid_modes[mi].num_rows;
    else
	rowheight = 16;
	
    startline = (BYTE)(y % rowheight);
    
    startaddr = (y / rowheight) * scrnwidth;
    startaddr += page * valid_modes[mi].pagesize;
    startaddr += x / charwidth;
    startcol = (BYTE)(x % charwidth);
    if (charwidth == 9)
	if (startcol == 0)
	    startcol = 8;
	else
	    --startcol;
    
    set_start_addr(startaddr);
    write_crtc_reg(PRESET_ROW_SCAN, startline);
    write_attrib_reg(H_PEL_PANNING, startcol);
}

/**** t e s t _ s p l i t _ s c r e e n ****
    Abstract:   Tests the split screen functions available on the
                VGA for one particular mode.  It draws a pattern on
                all available video pages, then scrolls the screen down,
                sideways, back, then up again.
    Parameters: The current mode in use.
    Returns:    Nothing
****************************/
void test_split_screen(int mi)
{
    int x, y;
    int a, b;
    BYTE page;
    
    SetMode (valid_modes[mi].video_mode);
    for (page=0; page<valid_modes[mi].max_pages; page++)
    {
	if (valid_modes[mi].isgraphics)
	{    
	    for (y = 0; y < valid_modes[mi].num_ypixels; y++)
	    {	 
		for (x = 0; x < valid_modes[mi].num_xpixels; x += y+1)
		{
		    SetPixel(page, 
			    ((page+4) % (valid_modes[mi].num_colors-1))+1,
			    x, y);
		}
	    }
	}
	writestring(NORMAL, page, page, page, "Mode %02Xh, page %d",
		valid_modes[mi].video_mode, page);
    }
    pausekey(2);

    set_pan_compat();
    SetPage(1);
    for (y = 0; y < 150; y++)
    {
	wait_for_vertical(0);
	set_line_compare(y);
	pausekey(0);
    }

    for (x = 0; x < y; x++)
    {
	wait_for_vertical(0);
	set_scrn_org(mi, 1, x, 0);
	pausekey(0); 
    }
    
    if (valid_modes[mi].isgraphics)
    {
	for (page=0; page<valid_modes[mi].max_pages; page++)
	{
	    for (a = 0; a < valid_modes[mi].num_ypixels; a++)
	    {
		for (b = 0; b < valid_modes[mi].num_xpixels; b += 16)
		{
		    SetPixel(page, 
			    ((page+2) % (valid_modes[mi].num_colors-1))+1,
			    b, a);
		}
	    }
	}
    }
    
    for (; x >= 0; x--) 
    {
	wait_for_vertical(0);
        set_scrn_org(mi, 1, x, 0);
        if (y)
            set_line_compare(--y);
	pausekey(0);
    }
     
    reset_pan_compat();
    set_line_compare(0x3FF);                /* maximum value */
    set_scrn_org(mi, 0, 0, 0);
}

/**** t e s t _ a t t r i b u t e _ r e g s ****
    Abstract:	    
		Palette Registers:  Try all palette values in 16 regs.
		Mode Control:  Palette size select (6+2 or 4+4) + Color Sel
		    Ignore PEL clock sel,  Graphics/ANUM
		    Try PEL Panning
		    Inten/Blink
		    Line Char/Turn off, get holes (char 219)
		    Select Display Type
		Overscan Color
		Color Plane enable
		Horizontal PEL Panning
                Test panning and scrolling of CRTC registers:
		    Preset Row Scan Register
		    Offset Register
		    Line Compare Register
    Parameters:
    Returns:
    Comments:
****************************/
void test_attribute_regs()
{
    GetStatus(&savemode);
    /*
    test_split_screen(COLORTEXT);
    pausekey(1 SEC);
    */
    if (vidmode == COLORVID)
	test_split_screen(COLOR16);
    else
	test_split_screen(MONOGRAPH);
    pausekey(1 SEC);
    SetMode(savemode.video_mode);
}
